import subprocess
import sys
import os
import psutil
import tempfile
import time
import shlex

def open_hide_terminal(command, with_qt=True, env=None):
    """
    Ouvre un nouveau terminal indépendant et exécute la commande spécifiée.
    attention command est une liste
    :param command: La liste de commande à exécuter.
    :param with_qt: Désactive l'affichage si False (utile pour les applications Qt sans affichage).
    :param hide_terminal: Masque la fenêtre du terminal si True.
    :return: Le PID du processus du terminal ouvert.
    """
    hide_terminal = True
    if env is None:
        env = dict(os.environ)
    if not with_qt:
        if sys.platform.startswith("Darwin") or sys.platform.startswith("darwin"):
            print("Warning: 'offscreen' mode may not be fully supported on Mac.")
        else:
            env['QT_QPA_PLATFORM'] = 'offscreen'
    if not with_qt:
        env['QT_QPA_PLATFORM'] = 'offscreen'

    process = None
    if not isinstance(command, list):
        raise RuntimeError("error command need to be a list")
    if len(command)==0:
        raise RuntimeError("error command need to be a list with elements")


    if sys.platform.startswith("win"):
        process = subprocess.Popen(
            command,
            env=env,
            shell=False,
            creationflags=(
                    subprocess.CREATE_NO_WINDOW |
                    subprocess.CREATE_NEW_PROCESS_GROUP
            ),
        )

        # startupinfo = subprocess.STARTUPINFO()
        # startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
        #
        # # Code 0 = SW_HIDE (cacher la console), mais ne gêne pas la GUI (ex : Qt)
        # startupinfo.wShowWindow = 0
        # creationflags = 0
        #
        # process = subprocess.Popen(
        #     command,
        #     env=env,
        #     startupinfo=startupinfo,
        #     creationflags=creationflags
        # )
    elif sys.platform.startswith("linux"):
        command_chaine = ' '.join(command)
        if hide_terminal:
            process = subprocess.Popen(["bash", "-c", command_chaine], env=env, stdout=subprocess.DEVNULL,
                                       stderr=subprocess.DEVNULL)
        else:
            try:
                process = subprocess.Popen(["gnome-terminal", "--", "bash", "-c", command_chaine + "; exec bash"], env=env)
            except FileNotFoundError:
                process = subprocess.Popen(["x-terminal-emulator", "-e", "bash", "-c", command_chaine + "; exec bash"],
                                           env=env)


    elif sys.platform.startswith("Darwin") or sys.platform.startswith("darwin") :
        command_chaine = ' '.join(command)
        if hide_terminal:
            if with_qt == False:
                command_chaine = "QT_QPA_PLATFORM=offscreen " + command_chaine
                # Lancer bash sans ouvrir un terminal visible
                process = subprocess.Popen(
                    ["/bin/bash", "-c", command_chaine],
                    env=env,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL
                )
            else:
                # Lancer bash sans ouvrir un terminal visible
                process = subprocess.Popen(
                    ["/bin/bash", "-c", command_chaine],
                    env=env,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL
                )
        else:
            if with_qt==False:
                command = "QT_QPA_PLATFORM=offscreen " + command_chaine
                # Sinon ouvrir Terminal.app
                process = subprocess.Popen(["osascript", "-e", f'tell app "Terminal" to do script \"{command_chaine}\"'])
            else:
                # Sinon ouvrir Terminal.app
                process = subprocess.Popen(["osascript", "-e", f'tell app "Terminal" to do script \"{command_chaine}\"'])
    if process:
        return process.pid
    else:
        raise RuntimeError("Impossible d'ouvrir le terminal.")


def open_terminal(command, with_qt=True, env=None):
    """
    attention command est une liste
    Ouvre un nouveau terminal indépendant et exécute la commande spécifiée.
    :param command: La liste de commande à exécuter.
    :param with_qt: Désactive l'affichage si False (utile pour les applications Qt sans affichage).
    :return: Le PID du processus du terminal ouvert.
    """
    if env is None:
        env = dict(os.environ)
    if not with_qt:
        if sys.platform.startswith("Darwin") or sys.platform.startswith("darwin"):
            print("Warning: 'offscreen' mode may not be fully supported on Mac.")
        else:
            env['QT_QPA_PLATFORM'] = 'offscreen'
    if not with_qt:
        env['QT_QPA_PLATFORM'] = 'offscreen'
    process = None
    if not isinstance(command, list):
        raise RuntimeError("error command need to be a list")
    if len(command)==0:
        raise RuntimeError("error command need to be a list with elements")



    if sys.platform.startswith("win"):
        # j insere cmd /k command
        # command.insert(0,"/k")
        # command.insert(0, "cmd.exe")
        # process = subprocess.Popen(command, env=env,creationflags=subprocess.CREATE_NEW_CONSOLE)
        process = subprocess.Popen(
            command,
            env=env,
            creationflags=(
                    subprocess.CREATE_NEW_CONSOLE |
                    subprocess.CREATE_NEW_PROCESS_GROUP
            )
        )

        if process:
            return process.pid
        else:
            raise RuntimeError("Impossible d'ouvrir le terminal.")
    elif sys.platform.startswith("linux"):
        command_chaine = ' '.join(command)
        try:
            process = subprocess.Popen(["gnome-terminal", "--", "bash", "-c", command_chaine + "; exec bash"], env=env)
        except FileNotFoundError:
            process = subprocess.Popen(["x-terminal-emulator", "-e", "bash", "-c", command_chaine + "; exec bash"], env=env)
        if process:
            return process.pid
        else:
            raise RuntimeError("Impossible d'ouvrir le terminal.")

    elif sys.platform.startswith("Darwin") or sys.platform.startswith("darwin"):
        # macOS: on écrit $$ (PID du shell) dans un fichier, puis on remplace le shell avec exec
        pidfile = tempfile.NamedTemporaryFile(delete=False)
        pidfile_path = pidfile.name
        pidfile.close()

        command_chaine = ' '.join(shlex.quote(c) for c in command)
        if not with_qt:
            command_chaine = f"QT_QPA_PLATFORM=offscreen {command_chaine}"

        # Important: "exec" pour que le shell soit remplacé par le vrai process
        wrapped_command = f'echo $$ > {pidfile_path}; exec {command_chaine}'

        applescript = f'''
                    tell application "Terminal" 
                        do script "bash -c \\"{wrapped_command}\\"" 
                    end tell'''
        osa_proc = subprocess.Popen(["osascript", "-e", applescript])
        osa_proc.wait()  # évite les zombies

        # Attente active pour que le PID soit écrit
        pid = None
        for _ in range(30):  # max 3 secondes
            try:
                with open(pidfile_path, "r") as f:
                    pid_str = f.read().strip()
                    if pid_str:
                        pid = int(pid_str)
                        break
            except FileNotFoundError:
                pass
            time.sleep(0.1)

        os.unlink(pidfile_path)  # Nettoyage
        if pid:
            return pid
        else:
            raise RuntimeError("Impossible de récupérer le PID du processus lancé dans Terminal.")
    else:
        raise RuntimeError("Unsupported platform.")


def execute_command(command: str, hidden: bool = True,env=None):
    """
    Exécute une commande dans un terminal caché ou visible avec option d'attente.

    :param command: La ligne de commande à exécuter
    :param hidden: Exécuter dans un terminal caché si True, visible sinon
    :param wait: Attendre la fin de l'exécution si True, sinon continuer en parallèle
    :return: (code de retour, PID du processus ou None en cas d'échec)
    """
    try:
        startupinfo = None
        creationflags = 0
        shell = False

        # Gestion spécifique pour Windows
        if os.name == "nt":
            if hidden:
                startupinfo = subprocess.STARTUPINFO()
                startupinfo.dwFlags |= subprocess.STARTF_USESHOWWINDOW
                creationflags = subprocess.CREATE_NEW_CONSOLE
            else:
                command = f'start cmd.exe /k echo {command} & {command}'
            shell = True  # Nécessaire pour certaines commandes Windows
        if env is None:
            # Détermine la manière de lancer le processus
            process = subprocess.Popen(
                command,
                shell=shell,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
                startupinfo=startupinfo if os.name == "nt" else None,
                creationflags=creationflags if os.name == "nt" else 0,
            )
        else:
            # Détermine la manière de lancer le processus
            process = subprocess.Popen(
                command,
                shell=shell,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                stdin=subprocess.PIPE,
                startupinfo=startupinfo if os.name == "nt" else None,
                creationflags=creationflags if os.name == "nt" else 0,
                env=env,
            )



        return 0, process.pid  # Succès avec PID

    except Exception as e:
        print(f"Erreur: {e}", file=sys.stderr)
        return 1, None  # Erreur


def kill_process(pid: int):
    """
    Tue un processus à partir de son PID.
    :param pid: L'identifiant du processus à tuer
    :return: 0 si succès, 1 en cas d'erreur
    """
    try:
        if os.name == "nt":
            subprocess.run(["taskkill", "/PID", str(pid), "/F"], check=True)
        else:
            os.kill(pid, 9)  # Signal SIGKILL
        return 0
    except Exception as e:
        print(f"Erreur lors de la suppression du processus {pid}: {e}", file=sys.stderr)
        return 1

def kill_process_tree(pid):
    try:
        parent = psutil.Process(pid)
        children = parent.children(recursive=True)
        for child in children:
            print(f"Tuer enfant PID {child.pid} ({child.name()})")
            child.kill()

        if sys.platform.startswith("Darwin") or sys.platform.startswith("darwin"):
            # Attendre un peu pour que les enfants meurent proprement
            gone, alive = psutil.wait_procs(children, timeout=3)
            for p in alive:
                print(f"Enfant PID {p.pid} n'a pas terminé, on le force")
                p.kill()

        parent.kill()
        print(f"Processus principal {pid} tué.")

    except psutil.NoSuchProcess:
        print(f"Le processus {pid} n'existe pas.")
    except Exception as e:
        print(f"Erreur lors du kill du process tree : {e}")



import copy
import os
import re
try:
    import GPUtil #sometimes errors occurs on gpu testing
except:
    pass
import psutil
import base64
import ntpath
import platform
from llama_cpp import Llama
try:
    from llama_cpp.llama_chat_format import Qwen3VLChatHandler
except Exception as e:
    print("Missing installs for Qwen3-VL:", e)
    pass
from Orange.data import Domain, StringVariable, Table


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.llm import prompt_management
    from Orange.widgets.orangecontrib.AAIT.utils import MetManagement
else:
    from orangecontrib.AAIT.llm import prompt_management
    from orangecontrib.AAIT.utils import MetManagement

def check_gpu(model_path, argself):
    """
    Checks if the GPU has enough VRAM to load a model.

    Args:
        model_path (str): Path to the model file.
        argself (OWWidget): OWQueryLLM object.

    Returns:
        bool: True if the model can be loaded on the GPU, False otherwise.
    """
    argself.error("")
    argself.warning("")
    argself.information("")
    argself.can_run = True
    argself.use_gpu = True
    return

    # attention bien faire la suite dans un try exept car gputil peut etre capricieux
    token_weight = 0.13
    if model_path is None:
        argself.use_gpu = False
        return
    if platform.system() != "Windows":
        argself.use_gpu = False
        return
    if not model_path.endswith(".gguf"):
        argself.use_gpu = False
        argself.can_run = False
        argself.error("Model is not compatible. It must be a .gguf format.")
        return
    # Calculate the model size in MB with a 1500 MB buffer
    model_size = os.path.getsize(model_path) / (1024 ** 3) * 1000
    model_size += token_weight * int(argself.n_ctx)
    print(f"Required memory: {model_size/1000:.2f}GB")
    # If there is no GPU, set use_gpu to False
    if len(GPUtil.getGPUs()) == 0:
        argself.use_gpu = False
        argself.information("Running on CPU. No GPU detected.")
        return
    # Else
    else:
        # Get the available VRAM on the first GPU
        gpu = GPUtil.getGPUs()[0]
        free_vram = gpu.memoryFree
    # If there is not enough VRAM on GPU
    if free_vram < model_size:
        # Set use_gpu to False
        argself.use_gpu = False
        # Check for available RAM
        available_ram = psutil.virtual_memory().available / 1024 / 1024
        if available_ram < model_size:
            argself.can_run = False
            argself.error(f"Cannot run. Both GPU and CPU are too small for this model (required: {model_size/1000:.2f}GB).")
            return
        else:
            argself.warning(f"Running on CPU. GPU seems to be too small for this model (available: {free_vram/1000:.2f}GB || required: {model_size/1000:.2f}GB).")
            return
    # If there is enough space on GPU
    else:
        try:
            # Load the model and test it
            # model = GPT4All(model_name=model_path, model_path=model_path, n_ctx=int(argself.n_ctx),
            #                 allow_download=False, device="cuda")
            # answer = model.generate("What if ?", max_tokens=3)
            # # If it works, set use_gpu to True
            argself.use_gpu = True
            argself.information("Running on GPU.")
            return
        # If importing Llama and reading the model doesn't work
        except Exception as e:
            # Set use_gpu to False
            argself.use_gpu = False
            argself.warning(f"GPU cannot be used. (detail: {e})")
            return


def count_tokens(model: Llama, text: str) -> int:
    """
    Count the number of tokens in a text, for a chosen model.
    """
    tokens = model.tokenize(text.encode("utf-8"))
    return len(tokens)


def load_model(model_path, use_gpu, n_ctx=10000, k_cache=0, v_cache=0):
    """
    Charge un modèle GGUF avec llama_cpp.Llama.

    - use_gpu=True : tente d'utiliser l'accélération (Metal/CUDA/Vulkan selon build)
      en mettant n_gpu_layers à -1 (= toutes les couches si possible).
    - use_gpu=False : CPU only (n_gpu_layers=0).
    """
    if not os.path.exists(model_path):
        print(f"Model could not be found: {model_path} does not exist")
        return

    try:
        # n_gpu_layers : -1 = toutes les couches si le binaire a un backend GPU (Metal/CUDA/Vulkan)
        n_gpu_layers = -1 if use_gpu else 0

        # n_threads : par défaut tous les cœurs logiques dispo moins 1 (pour avoir l'interface graphique qui ne freeze pas)
        n_threads = max(1, (os.cpu_count()-1 or 1))

        # NOTE : llama_cpp utilise n_ctx pour la taille de contexte
        model = Llama(
            model_path=model_path,
            n_ctx=n_ctx,
            n_threads=n_threads,
            n_gpu_layers=n_gpu_layers,
            # Quelques réglages sûrs
            use_mmap=True,
            use_mlock=False,
            embedding=False,
            verbose=True, #hahahahahaha remettre a False
            type_k=k_cache,
            type_v=v_cache
        )
        return model
    except Exception as e:
        print("Failed to load model with llama_cpp:", e)
        return


def generate_answers(table, model_path, use_gpu=False, n_ctx=4096, query_parameters=None, workflow_id="", progress_callback=None, argself=None):
    """
    Identique en signature/comportement, mais utilise llama_cpp sous le capot.
    """
    # Copie des données d'entrée
    data = copy.deepcopy(table)
    attr_dom = list(data.domain.attributes)
    metas_dom = list(data.domain.metas)
    class_dom = list(data.domain.class_vars)

    # Chargement modèle (llama_cpp)
    if not "Qwen3-VL-8B-Instruct" in model_path:
        model = load_model(model_path=model_path,
                           use_gpu=use_gpu,
                           n_ctx=n_ctx,
                           k_cache=query_parameters["k_cache"],
                           v_cache=query_parameters["v_cache"])
    else:
        model = load_Qwen3VL(model_path=model_path, n_ctx=n_ctx)
    if model is None:
        return None
    # Paramètres de génération par défaut
    if query_parameters is None:
        query_parameters = {"max_tokens": 4096, "temperature": 0.4, "top_p": 0.4, "top_k": 40, "repeat_penalty": 1.15}

    # Génération sur la colonne "prompt"
    try:
        rows = []
        for i, row in enumerate(data):
            features = list(data[i])
            metas = list(data.metas[i])
            prompt = row["prompt"].value

            system_prompt = row["system prompt"].value if "system prompt" in data.domain else ""
            assistant_prompt = row["assistant prompt"].value if "assistant prompt" in data.domain else ""

            # Appliquer ton template existant (inchangé)
            if not "Qwen3-VL-8B-Instruct" in model_path:
                prompt = prompt_management.apply_prompt_template(
                    model_path,
                    user_prompt=prompt,
                    assistant_prompt=assistant_prompt,
                    system_prompt=system_prompt
                )

                prompt = handle_context_length(prompt, model, n_ctx, method="truncate", margin=query_parameters["max_tokens"], progress_callback=progress_callback)

                answer = run_query(
                    prompt,
                    model=model,
                    max_tokens=query_parameters["max_tokens"],
                    temperature=query_parameters["temperature"],
                    top_p=query_parameters["top_p"],
                    top_k=query_parameters["top_k"],
                    repeat_penalty=query_parameters["repeat_penalty"],
                    workflow_id=workflow_id,
                    argself=argself,
                    progress_callback=progress_callback
                )
            else:
                image_paths = [p.strip() for p in row["image paths"].value.split(";")] if "image paths" in data.domain else []
                answer = run_Qwen3VL_query(query=prompt,
                                           image_paths=image_paths,
                                           model=model,
                                           progress_callback=progress_callback)

            if answer == "":
                answer = (
                    "Error: The answer could not be generated. Your prompt might be too long, or the model architecture you tried to use is possibly "
                    f"not supported yet.\n\nModel name: {ntpath.basename(model_path)}"
                )

            thinking = ""
            matches = re.findall(r"<think>[\s\S]*?</think>", answer)
            if matches:
                thinking = matches[0]
                answer = answer.replace(thinking, "").strip()
            metas += [answer, thinking]
            rows.append(features + metas)

            if progress_callback is not None:
                progress_value = float(100 * (i + 1) / len(data))
                progress_callback(("progressBar", progress_value))

            if argself is not None and getattr(argself, "stop", False):
                break
    except Exception as e:
        print("An error occurred when trying to generate an answer:", e)
        return

    # Ajouter la colonne "Answer" en metas
    answer_dom = [StringVariable("Answer"), StringVariable("Thinking")]

    domain = Domain(attributes=attr_dom, metas=metas_dom + answer_dom, class_vars=class_dom)
    out_data = Table.from_list(domain=domain, rows=rows)
    return out_data


class StopCallback:
    def __init__(self, stop_sequences, widget_thread=None):
        self.stop_sequences = stop_sequences
        self.recent_tokens = ""
        self.returning = True  # Store the last valid token before stopping
        self.widget_thread = widget_thread

    def __call__(self, token_id, token):
        # Stop in case thread is stopped
        if self.widget_thread:
            if self.widget_thread.stop:
                return False

        # Stop in case stop word has been met
        if not self.returning:
            return False
        self.recent_tokens += token

        # Check if any stop sequence appears
        for stop_seq in self.stop_sequences:
            if stop_seq in self.recent_tokens:
                self.returning = False  # Stop the generation, but allow the last token

        return True  # Continue generation


def write_tokens_to_file(token: str, workflow_id=""):
    chemin_dossier = MetManagement.get_api_local_folder(workflow_id=workflow_id)
    if os.path.exists(chemin_dossier):
        MetManagement.write_file_time(chemin_dossier + "time.txt")
        filepath = os.path.join(chemin_dossier, "chat_output.txt")
        with open(filepath, "a", encoding="utf-8") as f:
            f.write(token)
            f.flush()


def run_query(prompt, model, max_tokens=4096, temperature=0, top_p=0, top_k=40, repeat_penalty=1.15,
              workflow_id="", argself=None, progress_callback=None):
    """
    Version llama_cpp avec streaming.
    On garde la même signature et le même contrat de retour.
    """


    # Séquences d'arrêt à filtrer du résultat final
    stop_sequences = ["<|endoftext|>", "### User", "<|im_end|>", "<|im_start|>", "<|im_end>", "<im_end|>", "<im_end>"]
    callback_instance = StopCallback(stop_sequences, argself)

    # Paramètres de sampling mappés vers llama_cpp
    gen_kwargs = dict(
        max_tokens=max_tokens,
        temperature=temperature,
        top_p=top_p if top_p else 1.0,   # top_p=0 désactive → on met 1.0
        top_k=top_k if top_k else 0,     # top_k=0 désactive
        repeat_penalty=repeat_penalty,
        stream=True,
    )

    answer = ""

    # IMPORTANT :
    # - On utilise create_completion (prompt-style) pour rester compatible avec ton templating actuel.
    # - Le générateur renvoie des chunks contenant choices[0].text.
    try:
        stream = model(prompt=prompt, **gen_kwargs)

        for chunk in stream:
            # Récupérer le texte incrémental
            token = chunk["choices"][0].get("text", "")
            if not token:
                continue

            # Callback d'arrêt custom (on simule token_id=None)
            if not callback_instance(None, token):
                # On stoppe proprement le flux (consommation du générateur non nécessaire)
                answer += token  # on peut inclure le dernier token si souhaité
                break

            answer += token
            write_tokens_to_file(token, workflow_id)

            if progress_callback is not None:
                progress_callback(("assistant", token))

            if argself is not None and getattr(argself, "stop", False):
                # Arrêt demandé de l'extérieur
                return answer

    except Exception as e:
        # En cas d'erreur pendant la génération, on retourne ce qu'on a + log
        print("Generation error (llama_cpp):", e)

    # Nettoyage des séquences d'arrêt
    for stop in stop_sequences:
        if stop:
            answer = answer.replace(stop, "")

    return answer


def generate_conversation(table, model, conversation=None, n_ctx=32768, query_parameters=None, workflow_id="", progress_callback=None, argself=None):
    """
    Generates a response using a language model and appends it to a conversation.

    Parameters:
    ----------
    table (Orange.data.Table) Input data table. The first row should contain at least a "prompt" column, and optionally
        "system prompt" and "assistant prompt" columns for context.
    model (Llama) : Loaded language model instance, compatible with GPT4All or llama.cpp-style interfaces.
    widget (OWConverseLLM) : The widget in which to display the conversation.
    conversation (list, optional) : Existing conversation string to append the new response to. Defaults to an empty string.
    progress_callback (callable, optional) : Callback function for UI updates during generation. Called with progress percentage and message.
    argself (object, optional) : Extra argument passed to `run_query`, typically the widget instance (used for context or settings).

    Returns:
    -------
    Orange.data.Table
        A new Orange Table containing the original input row with two new meta columns:
        - "Answer": the model's generated response.
        - "Conversation": the updated full conversation string.
    """
    if table is None:
        return

    # Copy the input data
    data = copy.deepcopy(table)
    L = len(data)

    # Get data from the first row of the Data Table
    system_prompt = data[0]["system prompt"].value if "system prompt" in data.domain else ""
    user_prompt = data[0]["prompt"].value
    assistant_prompt = data[0]["assistant prompt"].value if "assistant prompt" in data.domain else ""

    # Get model path to determine prompt template
    model_path = model.model_path

    prompt = ""
    # If the conversation has not started, add a default system prompt
    if not conversation:
        conversation = [{"role": "system", "content": system_prompt}]

    # If the conversation has started, build the complete context
    for message in conversation:
        if message["role"] == "system":
            prompt += prompt_management.apply_system_template(model_path, message["content"])
        elif message["role"] == "user":
            prompt += prompt_management.apply_user_template(model_path, message["content"])
        elif message["role"] == "assistant":
            prompt += prompt_management.apply_assistant_template(model_path, message["content"])

    # Add the current user prompt & assistant prompt
    prompt += prompt_management.apply_user_template(model_path, user_prompt)
    prompt += prompt_management.apply_assistant_template(model_path, assistant_prompt)

    prompt = handle_context_length(prompt, model, n_ctx, method="truncate", margin=query_parameters["max_tokens"], progress_callback=progress_callback)

    # Return progression to fill the user & assistant cards
    progress_callback(("user", user_prompt))
    progress_callback(("assistant", assistant_prompt))

    # Append the user message to the conversation
    conversation.append({"role": "user", "content": user_prompt})

    # Generate the answer
    answer = run_query(prompt,
                       model=model,
                       max_tokens=query_parameters["max_tokens"],
                       temperature=query_parameters["temperature"],
                       top_p=query_parameters["top_p"],
                       top_k=query_parameters["top_k"],
                       repeat_penalty=query_parameters["repeat_penalty"],
                       workflow_id=workflow_id,
                       argself=argself,
                       progress_callback=progress_callback)

    # Split the thinking block and remove it from the answer
    think_text, final_answer = split_think(answer=answer)

    # Append the answer to the conversation
    conversation.append({"role": "assistant", "content": assistant_prompt + final_answer})

    # End of the progress bar
    progress_callback(("progressBar", 100))

    # Padding to add columns to the table, TEMPORARY # TODO : process the entire table for conversation
    rows_answer = [final_answer]
    rows_answer += [""] * (L - len(rows_answer))
    rows_think = [think_text]
    rows_think += [""] * (L - len(rows_think))
    conv_text = ""
    for message in conversation:
        if message["role"] == "system":
            conv_text += f"# System:\n{message['content']}\n\n"
        elif message["role"] == "user":  # Changé en elif
            conv_text += f"# User:\n{message['content']}\n\n"
        elif message["role"] == "assistant":
            conv_text += f"# Assistant:\n{message['content']}\n\n"
    rows_conv = [conv_text]
    rows_conv += [""] * (L - len(rows_conv))

    # Generate new Domain to add to data
    var_answer = StringVariable("Answer")
    var_conversation = StringVariable("Conversation")
    var_think = StringVariable("Think")

    # Add columns to the table
    data = data.add_column(var_answer, rows_answer, to_metas=True)
    data = data.add_column(var_conversation, rows_conv, to_metas=True)
    data = data.add_column(var_think, rows_think, to_metas=True)

    return data, conversation


def split_think(answer: str):
    # Extract think content (if any)
    think_match = re.search(r"<think>(.*?)</think>", answer, flags=re.DOTALL)
    think_text = think_match.group(1).strip() if think_match else ""
    # Remove think block from the final answer
    final_answer = re.sub(r"<think>.*?</think>", "", answer, flags=re.DOTALL).strip()
    return think_text, final_answer


def handle_context_length(prompt, model, n_ctx, method="truncate", margin=0, progress_callback=None):
    """
    Truncate a prompt to fit within n_ctx tokens, leaving margin for generation.
    Safely handles edge cases where limit <= 0.
    """
    # Keep a margin for generated tokens
    limit = max(n_ctx - margin, 0)  # clamp to at least 0

    if method == "truncate":
        tokens = model.tokenize(prompt.encode("utf-8"))  # pass string, not bytes
        initial_length = len(tokens)
        if initial_length > limit:
            # take last `limit` tokens safely
            tokens = tokens[-limit:] if limit > 0 else []
            truncated_length = len(tokens)
            prompt = model.detokenize(tokens).decode("utf-8") if tokens else ""
            if progress_callback:
                warning = (
                    f"Complete prompt contains {initial_length} tokens - context limit is {limit} (Context length - Max tokens). "
                    f"The {truncated_length} last tokens have been kept in the prompt."
                )
                progress_callback(("warning", warning))
        return prompt
    elif method == "summarize":
        pass
    else:
        return prompt



# For pure display
def conversation_to_text(conversation):
    if not conversation:
        return ""
    else:
        pass #TODO



def load_Qwen3VL(model_path, n_ctx=32768):
    folder = os.path.dirname(model_path)
    mmproj_path = os.path.join(folder, "mmproj-Qwen3-VL-8B-Instruct-F16.gguf")
    if not os.path.exists(mmproj_path):
        print("Something is wrong with your model : couldn't find mmproj-Qwen3-VL-8B-Instruct-F16.gguf")
        return None
    chat_handler = Qwen3VLChatHandler(clip_model_path=mmproj_path)
    model = Llama(model_path=model_path,
                  chat_handler=chat_handler,
                  n_ctx=n_ctx,
                  n_gpu_layers=-1)
    return model


def run_Qwen3VL_query(query, image_paths, model, system_prompt=" ", progress_callback=None):
    image_messages = []
    for image_path in image_paths:
        data_uri = convert_to_uri(image_path)
        if not data_uri.startswith("data"):
            progress_callback(("error", data_uri))
            return "The image could not be processed"
        image_messages.append({"type": "image_url", "image_url": {"url": data_uri}})
    image_messages.append({"type": "text", "text": query})

    messages = [{"role": "system", "content": system_prompt},
                {"role": "user", "content": image_messages}]
    generator = model.create_chat_completion(messages=messages, stream=True)

    full_response = ""
    for chunk in generator:
        # chunk is a dict, often with a 'choices' list
        for choice in chunk.get("choices", []):
            # Each choice may have a 'delta' dict with 'content'
            delta = choice.get("delta", {})
            token = delta.get("content")
            if token:
                full_response += token
                if progress_callback is not None:
                    progress_callback(("assistant", token))
    return full_response



_IMAGE_MIME_TYPES = {
    # Most common formats
    '.png':  'image/png',
    '.jpg':  'image/jpeg',
    '.jpeg': 'image/jpeg',
    '.gif':  'image/gif',
    '.webp': 'image/webp',
    '.svg':  'image/svg+xml',
    '.svgz': 'image/svg+xml',

    # Next-generation formats
    '.avif': 'image/avif',
    '.heic': 'image/heic',
    '.heif': 'image/heif',
    '.heics': 'image/heic-sequence',
    '.heifs': 'image/heif-sequence',

    # Legacy / Windows formats
    '.bmp':  'image/bmp',
    '.dib':  'image/bmp',
    '.ico':  'image/x-icon',
    '.cur':  'image/x-icon',

    # Professional imaging
    '.tif':  'image/tiff',
    '.tiff': 'image/tiff',
}

def convert_to_uri(
    file_path: str,
    fallback_mime: str = "image/png" #"application/octet-stream"
) -> str:
    """
    Convert a local image file to a base64-encoded data URI with the correct MIME type.

    Supports 20+ image formats (PNG, JPEG, WebP, AVIF, HEIC, SVG, BMP, ICO, TIFF, etc.).

    Args:
        file_path: Path to the image file on disk.
        fallback_mime: MIME type used when the file extension is unknown.

    Returns:
        A valid data URI string (e.g., data:image/webp;base64,...).

    Raises:
        FileNotFoundError: If the file does not exist.
        OSError: If reading the file fails.
    """
    if not os.path.isfile(file_path):
        return f"Image file not found: {file_path}"

    extension = os.path.splitext(file_path)[1].lower()
    mime_type = _IMAGE_MIME_TYPES.get(extension, fallback_mime)

    if mime_type == fallback_mime:
        print(f"Warning: Unknown extension '{extension}' for '{file_path}'. "
              f"Using fallback MIME type: {fallback_mime}")

    try:
        with open(file_path, "rb") as img_file:
            encoded_data = base64.b64encode(img_file.read()).decode("utf-8")
    except Exception as e:
        return f"Failed to read image file '{file_path}': {e}"

    return f"data:{mime_type};base64,{encoded_data}"





import copy
import spacy
from spacy.tokenizer import Tokenizer
from spacy.util import compile_infix_regex
from spacy.symbols import ORTH
# import re
from unidecode import unidecode

from Orange.data import StringVariable, DiscreteVariable, Domain, Table

pos_tags = ["ADJ", "ADP", "ADV", "AUX", "CCONJ", "DET", "INTJ", "NOUN", "NUM", "PART", "PRON", "PROPN", "PUNCT",
            "SCONJ", "SYM", "VERB", "X", "REF"]


def custom_tokenizer(nlp):
    tokenizer = Tokenizer(nlp.vocab)

    # Add custom rule to avoid splitting words that contain numbers
    # This pattern matches any word that contains at least one number (with optional symbols)
    tokenizer.add_special_case(r'\b[A-Za-z0-9#.-]+\b', [{ORTH: "@"}])  # Treat these as a single token

    return tokenizer


def create_lemmes_and_tags(table, model_path, request=False, progress_callback=None, argself=None):
    if request:
        out_data = create_lemmes_and_tags_on_request(table, model_path)
    else:
        out_data = create_lemmes_and_tags_on_text(table, model_path, progress_callback, argself)
    return out_data


def create_lemmes_and_tags_on_text(table, model_path, progress_callback=None, argself=None):
    """
    Add lemmes and tags columns to an input Data Table.

    Parameters:
    table (Table): The input Table to process.
    model_path (str): The path to the NLP model.

    Returns:
    out_data (Table): Copy of the input Table with 2 additional columns - lemmes and tags
    """
    # Copy of input data
    data = copy.deepcopy(table)
    attr_dom = list(data.domain.attributes)
    metas_dom = list(data.domain.metas)
    class_dom = list(data.domain.class_vars)
    # Load the model
    model = spacy.load(model_path)
    # Generate lemmes & tags on column named "content"
    rows = []
    for i, row in enumerate(data):
        features = list(data[i])
        metas = list(data.metas[i])
        lemmes, tags = lemmatize(str(row["content"]), model)
        lemmes = [" ".join(lemmes)]
        tags = [" ".join(tags)]
        metas += lemmes
        metas += tags
        rows.append(features + metas)
        if progress_callback is not None:
            progress_value = float(100 * (i + 1) / len(data))
            progress_callback(progress_value)
        if argself is not None:
            if argself.stop:
                break
    # Generate new Domain to add to data
    var_lemmes = StringVariable(name="Lemmes")
    var_tags = StringVariable(name="Tags")
    domain = Domain(attributes=attr_dom, metas=metas_dom + [var_lemmes, var_tags], class_vars=class_dom)
    # Create and return table
    out_data = Table.from_list(domain=domain, rows=rows)
    return out_data


def create_lemmes_and_tags_on_request(table, model_path):
    # Assume the input Table contains only one row for the request
    request = table[0]["content"].value
    # Load the model
    model = spacy.load(model_path)
    # Generate lemmes & tags
    lemmes, tags = lemmatize(request, model)
    # Rearrange lemmes and tags to fit in a Table
    rows = [[tags[i], lemmes[i]] for i in range(len(lemmes))]
    # Create domain
    var_lemmes = StringVariable(name="Lemmes")
    var_tags = DiscreteVariable(name="Tags", values=pos_tags)
    domain = Domain(attributes=[var_tags], metas=[var_lemmes])
    # Create and return Table
    out_data = Table.from_list(domain=domain, rows=rows)
    return out_data


def normalize_text(text):
    """
    Normalize text by removing accents and converting to lowercase using `unidecode`.

    Args:
        text (str): Input text.

    Returns:
        str: Normalized text.
    """
    return unidecode(text).lower()  # Convert accents and lowercase


def lemmatize(text, model):
    """
    Computes the lemmas & tags of a text using a spaCy model.
    Reference numbers (tokens with digits) are tagged as 'REF' and kept as-is.

    Parameters:
    text (str): The text to process.
    model (spacy.language.Language): The spaCy model to use for processing.

    Returns:
    tuple: Two lists containing the lemmas and tags of each token.
    """
    lemmes = []
    tags = []
    # Modify the infix patterns (patterns for token splits)
    infixes = list(model.Defaults.infixes)
    # Add custom pattern for numbers with special characters
    infixes.append(r"(?=.*[A-Za-z])(?=.*\d)[A-Za-z\d]+(?:[-.#][A-Za-z\d]+)*")
    infixes.append(r"[^\s]*")
    # Recompile the infix pattern after adding the custom one
    model.tokenizer.infix_finditer = compile_infix_regex(infixes).finditer
    document = model(text)
    for token in document:
        if any([c.isdigit() for c in token.text]):
            tags.append("REF")
            lemmes.append(token.text)  # Keep original text for reference numbers
        else:
            lemmes.append(normalize_text(token.lemma_))
            if token.pos_ not in pos_tags:
                tags.append("X")
            else:
                tags.append(token.pos_)
    return lemmes, tags




import os
import sys
import Orange.data
from Orange.data import Domain, Table,  ContinuousVariable
from Orange.widgets.utils.signals import Input
from Orange.widgets.widget import Output, OWWidget
import numpy as np
from Orange.widgets.settings import Setting
from AnyQt.QtWidgets import QCheckBox, QPushButton


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.unlink_table_domain import unlink_domain

else:
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.unlink_table_domain import unlink_domain

class LoopStartWidget(OWWidget):
    name = "Loop Start"
    description = "Widget to start a loop with data table input and output."
    icon = "icons/startloop.png"
    category = "AAIT - ALGORITHM"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/startloop.png"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owstartloop.ui")
    want_control_area = False
    priority = 1010
    str_iterate_still_nb_line_dont_change: str = Setting("False")
    str_iter_of_line_number: str = Setting("False")
    class Inputs:
        data_in = Input("Data In", Orange.data.Table)
        # in_pointer = Input("End of the Loop Do-While", str, auto_summary=False)

    class Outputs:
        data_out = Output("Data Out", Orange.data.Table)
        out_pointer = Output("Begin of the Loop Do-While", str, auto_summary=False)

    def __init__(self):
        super().__init__()

        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)
        self.checkBox_iteration_until_number_of_line_dont_change = self.findChild(QCheckBox, 'checkBox')
        self.checkBox_iteration_on_the_number_of_line = self.findChild(QCheckBox, 'checkBox_2')
        self.pushButton=self.findChild(QPushButton, 'pushButton')
        self.pushButton.clicked.connect(self.button_reintialise_action)

        if self.str_iterate_still_nb_line_dont_change !="False":
            self.checkBox_iteration_until_number_of_line_dont_change.setChecked(True)
        else:
            self.checkBox_iteration_until_number_of_line_dont_change.setChecked(False)
        self.checkBox_iteration_until_number_of_line_dont_change.stateChanged.connect(self.update_checkBox_iteration_until_number_of_line_dont_change)
        if self.str_iter_of_line_number !="False":
            self.checkBox_iteration_on_the_number_of_line.setChecked(True)
        else:
            self.checkBox_iteration_on_the_number_of_line.setChecked(False)
        self.checkBox_iteration_on_the_number_of_line.stateChanged.connect(self.updat_checkBox_iteration_on_the_number_of_line)
        self.iter = 0
        self.data = None


    def button_reintialise_action(self):
        self.iter = 0
        dataset = self.data
        self.set_data(dataset)
        self.send_pointer()

    def update_checkBox_iteration_until_number_of_line_dont_change(self):
        if self.checkBox_iteration_until_number_of_line_dont_change.isChecked():
            self.str_iterate_still_nb_line_dont_change="True"
            self.checkBox_iteration_on_the_number_of_line.setChecked(False)
        else:
            self.str_iterate_still_nb_line_dont_change = "False"

    def updat_checkBox_iteration_on_the_number_of_line(self):
        if self.checkBox_iteration_on_the_number_of_line.isChecked():
            self.str_iter_of_line_number="True"
            self.checkBox_iteration_until_number_of_line_dont_change.setChecked(False)
        else:
            self.str_iter_of_line_number = "False"


    @Inputs.data_in
    def set_data(self, dataset):
        if dataset is None:
            print("No data received.")
            return
        self.send_pointer()
        self.error("")
        self.data=None
        if self.str_iterate_still_nb_line_dont_change =="True":
            self.data=dataset
            self.process_data()
            return
        if self.str_iter_of_line_number =="True":
            self.data=dataset
            self.execute_iter_of_line_number()
            return
        error_code_check_iter_column=self.check_iter_column(dataset)
        if error_code_check_iter_column==1:
            self.data=self.add_iter_column(dataset)
        elif error_code_check_iter_column==0:
            self.data = dataset
        else:
            self.error("error remove iter column in input data!!!")
            return
        self.process_data()

    def add_iter_column(self,in_data_orange):
        iter_var = ContinuousVariable("iter")
        new_domain = Domain(in_data_orange.domain.attributes + (iter_var,), in_data_orange.domain.class_var, in_data_orange.domain.metas)
        iter_values = np.ones((len(in_data_orange), 1))  # column of 0
        # merge
        new_X = np.hstack((in_data_orange.X, iter_values))
        return Table(new_domain, new_X, in_data_orange.Y, in_data_orange.metas)

    def check_iter_column(self,in_data_orange):
        """
        Checks the presence and validity of the categorical variable 'iter' in in_data.

        Returns:
        - 0 if 'iter' is a categorical variable with only "0" and "1" as values.
        - 1 if 'iter' is missing from the domain.
        - 2 if 'iter' exists but is either not continuous.
        """
        # Check if "iter" exists in the domain
        if "iter" not in in_data_orange.domain:
            return 1  # "iter" does not exist

        iter_var = in_data_orange.domain["iter"]

        # Check if "iter" is continuous
        if not iter_var.is_continuous:
            return 2  # "iter" is not continuous

        return 0  # Everything is correct


    def get_nb_line(self):
        """Return the number of lines to be called from another widget."""
        return 0 if self.data is None else len(self.data)

    def get_iter(self):
        return self.iter

    def reinitialize_iter(self):
        self.iter = 0
        return


    def is_allow_to_change_line_number(self):
        """Return if we are allowed to change the number of line to be called from another widget."""
        return self.str_iterate_still_nb_line_dont_change

    def iter_of_line_number(self):
        return self.str_iter_of_line_number

    def execute_iter_of_line_number(self):
        table = Table.from_table(domain=self.data.domain, source=self.data, row_indices=[self.iter])
        self.iter += 1
        self.Outputs.data_out.send(unlink_domain(table))

    def get_column_name_and_type(self):
        """Return the name and type of 'data_in' to be called from another widget."""
        if self.data is None:
            return [[], []]
        column_names = []
        column_types = []

        for element in self.data.domain.variables + self.data.domain.metas:
            column_names.append(str(element.name))
            column_types.append(str(type(element)))
        return column_names, column_types

    def get_in_data(self):
        return self.data


    def process_data(self):
        """Main process executed when data is available."""
        if self.data is not None:
            #new_data=copy.deepcopy(self.data)
            self.Outputs.data_out.send(unlink_domain(self.data))  # Envoie les données en sortie
        else:
            print("No data sent.")


    def send_pointer(self):
        """Send a pointer to the current class for the loop."""
        pointer = str(id(self))
        self.Outputs.out_pointer.send(pointer)

if __name__ == "__main__":
    from AnyQt.QtWidgets import QApplication

    app = QApplication(sys.argv)
    obj = LoopStartWidget()
    obj.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



# https://orange3.readthedocs.io/projects/orange-development/en/latest/tutorial-settings.html
import os
import sys

import Orange.data
from Orange.data import ContinuousVariable, Domain
from Orange.data.table import Table
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.optimiser import optuna_multi
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.optimiser import optuna_multi
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file


@apply_modification_from_python_file(filepath_original_widget=__file__)
class Optimisation(widget.OWWidget):
    name = "Optimization"
    description = "Optimization with optuna"
    icon = "icons/owoptimisation.svg"
    category = "AAIT - ALGORITHM"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owoptimisation.svg"



    priority = 1020
    dossier_du_script = os.path.dirname(os.path.abspath(__file__))
    want_control_area = False
    partie_1 = ""
    partie_2 = ""
    CaptionDuLabelPourGoto = ""
    AddAcondition = 0
    liste_a_afficher_comboBox_selection_label = []

    class Inputs:
        data_search_space = Input("Search Space Limits", Orange.data.Table)
        data_number_iterations = Input("Number of iterations", Orange.data.Table)
        data_previous_study = Input("Previous Study Data", Orange.data.Table)

    class Outputs:
        data_out = Output("Data", Orange.data.Table)
        best_trial_out = Output("Best trial", Orange.data.Table)
        current_proposition_out = Output("Current proposition (preview)", Orange.data.Table)
        studies_out = Output("Last Studies ", Orange.data.Table, auto_summary=False)
        out_pointer = Output("begin of the loop", str, auto_summary=False)  # sended as the widget is created
        out_pointer1 = Output("Send study for the loop", str, auto_summary=False)  # sended as the widget is created

    @Inputs.data_number_iterations
    def set_data_number_iterations(self, dataset):
        if dataset is not None:
            self.current_data_number_iterations = dataset[0][0]
            self.compteur_iteration = 1
            # pour relancer une etude
            self.itere(score=None)

    @Inputs.data_search_space
    def set_data_search_space(self, dataset):
        if dataset is not None:
            self.current_data_search_space = dataset
            self.len_file_current_data_search_space = len(dataset)

    @Inputs.data_previous_study
    def set_data_previous_study(self, dataset):
        self.current_data_previous_study = dataset


    # on lance l'étude et définition des data table en sortie
    def launch_study(self, score):
        ligne_best_etude, data, domain, current_proposition_out = optuna_multi.launch_study(score, self.len_score, self.current_data_search_space, self.current_data_previous_study, self.stop)

        if ligne_best_etude != []:
            table_best_etude = Table.from_list(Domain([ContinuousVariable(h) for h in domain]), ligne_best_etude)
            self.Outputs.best_trial_out.send(table_best_etude)

        if current_proposition_out != []:
            self.table = Table.from_list(Domain([ContinuousVariable(h) for h in domain]), current_proposition_out)
            self.Outputs.current_proposition_out.send(self.table)

        self.current_data_previous_study = Table.from_list(Domain([ContinuousVariable(h) for h in domain]), data[3:])
        self.Outputs.studies_out.send(self.current_data_previous_study)

        if self.stop == True:
            self.stop = False
            return

    def change_last_ligne(self, data=[]):
        if data == []:
            return

        if self.current_data_previous_study is not None:
            new_ligne = []
            for i in range(len(data.domain.attributes)):
                if data[0][i] == '?':
                    new_ligne.append("?")
                else:
                    # si le pas est défini dans le fichier d'entrée on gère le pas pour éviter toute cassure avec optuna
                    if len(self.current_data_search_space) == 3:
                        new_value = round(data[0][i].value / self.current_data_search_space[2][i].value) * self.current_data_search_space[2][i].value
                        data[0][i] = new_value
                    new_ligne.append(data[0][i].value)

            replace = False
            for i in range(len(data.domain.attributes)):
                if self.current_data_previous_study[-1][i] == '?':
                    replace = True
            #modification de la dernière ligne
            if replace == True:
                self.current_data_previous_study[-1] = new_ligne

            #ajout d'une nouvelle étude
            else:
                data = []
                for i in range(len(self.current_data_previous_study)):
                    d = []
                    for j in range(len(self.current_data_previous_study.domain)):
                        d.append(self.current_data_previous_study[i][j].value)

                    data.append(d)
                data.append(new_ligne)
                self.current_data_previous_study = Table.from_list(self.current_data_previous_study.domain,
                                                                    data)

            self.Outputs.studies_out.send(self.current_data_previous_study)

    # appelé depuis un focntion tierce
    def itere(self, score=[]):
        if self.current_data_search_space is None:
            print('Erreur il n y a pas de paramètres de configuration en entree')
            return

        if self.current_data_number_iterations is None:
            if self.compteur_iteration > 10:
                self.stop = True
        else:
            if self.compteur_iteration > self.current_data_number_iterations:
                self.stop = True

        self.compteur_iteration = self.compteur_iteration + 1

        # dans le cas du chargement d'une ancienne étude
        if self.current_data_previous_study is not None:
            self.len_score = len(self.current_data_previous_study.domain) - len(self.current_data_search_space.domain)

        if score is not None:
            new_score = []
            for i in range(len(score[0])):
                if score[0][i] != '?':
                   new_score.append(score[0][i].value)
            # ajouter car le feature constructor peut garder les anciens scores
            if self.len_score is None:
                self.len_score = len(new_score)

            #if len(new_score) != self.len_score:
                #print('Erreur les nombres de scores entre les études sont différents')
                #return
            self.launch_study(new_score)

        else:
            self.launch_study(score)



    def __init__(self):

        self.current_data_standard_input = None
        self.current_data_search_space = None
        self.current_data_previous_study = None
        self.current_data_number_iterations = None
        self.etude = None
        self.stop = False
        self.len_score = None
        self.table = None
        # compteur_iteration
        self.compteur_iteration = 1
        self.len_file_current_data_search_space = 0

        # on envoie l identifiant pour pouvoir boucler
        self.Outputs.out_pointer.send(str(id(self)))

        # on envoie l identifiant pour pouvoir boucler
        self.Outputs.out_pointer1.send(str(id(self)))

        super().__init__()  # je pense que c'est pour garder les config entre deux executions


        self.setFixedWidth(480)
        self.setFixedHeight(354)
        self.setAutoFillBackground(True)


if __name__ == "__main__":
    from AnyQt.QtWidgets import QApplication
    app = QApplication(sys.argv)
    mon_objet = Optimisation()
    mon_objet.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()






import os
import sys

import Orange.data
from AnyQt.QtWidgets import QApplication
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output

if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.llm import lemmes
    from Orange.widgets.orangecontrib.AAIT.utils import thread_management
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.llm import lemmes
    from orangecontrib.AAIT.utils import thread_management
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWLemmatizer(widget.OWWidget):
    name = "Lemmes & Tags"
    description = "Generate the lemmes and the tags on the column 'content' of a Table"
    category = "AAIT - LLM INTEGRATION"
    icon = "icons/owlemmatizer.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owlemmatizer.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owlemmatizer.ui")
    want_control_area = False
    priority = 1070

    class Inputs:
        request = Input("Request", Orange.data.Table)
        data = Input("Data", Orange.data.Table)
        model_path = Input("Model", str, auto_summary=False)

    class Outputs:
        data = Output("Data", Orange.data.Table)

    @Inputs.request
    def set_request(self, in_request):
        self.data = in_request
        self.request = True
        if self.autorun == True:
            self.run()

    @Inputs.data
    def set_data(self, in_data):
        self.data = in_data
        self.request = False
        if self.autorun==True:
            self.run()

    @Inputs.model_path
    def set_model(self, in_model_path):
        self.model_path = in_model_path
        if self.autorun == True:
            self.run()

    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(470)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)

        # Data Management
        self.request = False
        self.data = None
        self.model_path = None
        self.thread = None
        self.autorun = True
        self.result = None
        self.post_initialized()

    def run(self):
        # if thread is running quit
        if self.thread is not None:
            self.thread.safe_quit()

        if self.data is None:
            self.Outputs.data.send(None)
            return

        if self.model_path is None:
            self.Outputs.data.send(None)
            return

        # Verification of in_data
        self.error("")
        if not "content" in self.data.domain:
            self.error('You need a "content" column in input data')
            return

        if type(self.data.domain["content"]).__name__ != 'StringVariable':
            self.error('"content" column needs to be a Text')
            return

        # Start progress bar
        self.progressBarInit()

        # Connect and start thread : main function, progress, result and finish
        # --> progress is used in the main function to track progress (with a callback)
        # --> result is used to collect the result from main function
        # --> finish is just an empty signal to indicate that the thread is finished
        self.thread = thread_management.Thread(lemmes.create_lemmes_and_tags, self.data, self.model_path, self.request)
        self.thread.progress.connect(self.handle_progress)
        self.thread.result.connect(self.handle_result)
        self.thread.finish.connect(self.handle_finish)
        self.thread.start()

    def handle_progress(self, value: float) -> None:
        self.progressBarSet(value)

    def handle_result(self, result):
        try:
            self.result = result
            self.Outputs.data.send(result)
        except Exception as e:
            print("An error occurred when sending out_data:", e)
            self.Outputs.data.send(None)
            return

    def handle_finish(self):
        self.progressBarFinished()

    def post_initialized(self):
        pass

if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWLemmatizer()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



"""
Widgets from Development workflow category

"""
# ID = "orangecontrib.AAIT"

NAME = "Advanced Artificial Intelligence Tools"

# Category icon show in the menu
ICON = "icons/category.svg"

BACKGROUND = "light-green"

DESCRIPTION = ("Advanced Artificial Intelligence Tools is a package meant to develop and enable advanced AI "
               "functionalities in Orange Data Mining.")



# PRIORITY = 6


import os
import sys
import ntpath
import Orange.data
from AnyQt.QtWidgets import QApplication, QLabel
from Orange.widgets import widget
from Orange.widgets.utils.signals import Input, Output
from Orange.data import Domain, StringVariable, Table


if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
    from Orange.widgets.orangecontrib.AAIT.llm import lmstudio
else:
    from orangecontrib.AAIT.utils.import_uic import uic
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
    from orangecontrib.AAIT.llm import lmstudio

@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWLMStudio(widget.OWWidget):
    name = "LM Studio"
    description = "Query LM studio to get a response"
    icon = "icons/lm_studio.png"
    category = "AAIT - LLM INTEGRATION"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/lm_studio.png"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owllm4all.ui")
    want_control_area = False
    priority = 1110
    class Inputs:
        data = Input("Data", Orange.data.Table)
        model = Input("Model", str, auto_summary=False)

    class Outputs:
        data = Output("Data", Orange.data.Table)

    @Inputs.model
    def set_model(self, model):
        """
        Setter for the model.

        Parameters
        ----------
        model : str
            The model.

        Returns
        -------
        None
        """
        if not self.runable:
            return
        self.error("")
        # Set the model
        self.model = model
        # if not self.model.endswith(".gguf"):
        #     self.error("Model needs to be in a gguf format.")
        #     return
        if self.model is None:
            return
        if self.data is None:
            self.error("widget need an input data with a prompt column")
            return
        if "prompt" not in self.data.domain:
            self.error("input table need a prompt column")
            return
        # Run the widget
        self.run()

    @Inputs.data
    def set_data(self, in_data):
        """
        Setter for the input data.

        Parameters
        ----------
        in_data : Orange.data.Table
            The input table.

        Returns
        -------
        None

        """

        if not self.runable:
            return
        self.error("")
        # Set the input data
        self.data = in_data
        if self.data is None:
            return
        if "prompt" not in self.data.domain:
            self.error("input table need a prompt column")
            return
        if self.model==None:
            return
        # Run the widget
        self.run()

    def __init__(self):
        """
        Initialize the widget.

        This function initializes the widget and sets up its basic properties.
        It also loads the user interface file and finds the label for the description.
        """

        super().__init__()
        self.runable=True
        # Initialize path management
        # This is used to store the current Orange Widgets (OWS) path
        self.current_ows = ""

        # Set the fixed width and height of the widget
        self.setFixedWidth(470)
        self.setFixedHeight(300)

        # Load the user interface file
        uic.loadUi(self.gui, self)

        # Find the label for the description
        self.label_description = self.findChild(QLabel, 'Description')

        # Initialize data management
        # This is used to store the input data
        self.data = None
        self.model = None#"solar-10.7b-instruct-v1.0.Q6_K.gguf"
        self.result = None

    def run(self):
        """
        Run the widget.

        This function runs the widget by initializing the thread and starting it.
        It also handles the case where the thread is already running and interrupts it.

        Returns:
            None
        """
        # Clear the error message
        self.error("")

        # If data is not provided, exit the function
        if self.data is None:
            self.Outputs.data.send(None)
            return

        model_list = lmstudio.get_model_lmstudio()
        if isinstance(model_list, str) and model_list == "Error":
            self.error("Please launch the LMStudio app and start the server.")
            self.Outputs.data.send(None)
            return
        # Verify that the models directory is set in LMStudio (aait_store)
        if model_list is None:
            self.error("Your models directory in LMStudio is not set properly.")
            self.Outputs.data.send(None)
            return
        # Verify that the plugged model exists in your LMStudio model list
        if not any(ntpath.basename(d["id"]).lower() == ntpath.basename(self.model.lower()) for d in model_list["data"]):
            ids = [item["id"] for item in model_list.get("data", [])]
            print(
                f"Model not found in LMStudio. You are trying to use {ntpath.basename(self.model.lower())}, please verify that this model's API identifier exists in LMStudio.\n available model :{str(ids)}")
            self.error(
                f"Model not found in LMStudio. You are trying to use {ntpath.basename(self.model.lower())}, please verify that this model's API identifier exists in LMStudio.\n available model :{str(ids)}")
            self.Outputs.data.send(None)
            return
        for d in model_list["data"]:
            if ntpath.basename(d["id"]).lower() == ntpath.basename(self.model.lower()):
                self.model = d["id"]
        stream = False
        if "stream" in self.data.domain:
            stream = True
        rows = []
        for i, row in enumerate(self.data):
            features = list(self.data[i])
            metas = list(self.data.metas[i])
            result = lmstudio.appel_lmstudio(str(row["prompt"]), self.model, stream=stream)
            if result is None:
                self.error("Response generation error. See LM Studio logs.")
                self.Outputs.data.send(None)
                return
            metas.append(result)
            rows.append(features + metas)
        answer_dom = [StringVariable("Answer")]
        domain = Domain(attributes=self.data.domain.attributes, metas=(list(self.data.domain.metas)) + answer_dom,
                        class_vars=self.data.domain.class_vars)
        self.Outputs.data.send(Table.from_list(domain, rows=rows))



if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWLMStudio()
    my_widget.show()
    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os
import sys
import Orange.data
from AnyQt.QtWidgets import QApplication, QLineEdit, QCheckBox,QSpinBox, QPushButton
from Orange.data import StringVariable, Table, Domain
from Orange.widgets.utils.signals import Input, Output
from Orange.widgets.settings import Setting
from datetime import datetime, timedelta
from pathlib import Path
import numpy as np
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils import base_widget
    from Orange.widgets.orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file
else:
    from orangecontrib.AAIT.utils import base_widget
    from orangecontrib.AAIT.utils.initialize_from_ini import apply_modification_from_python_file


@apply_modification_from_python_file(filepath_original_widget=__file__)
class OWFileFromDir(base_widget.BaseListWidget):
    name = "Find Files From Dir"
    description = ("Search files by extension or no for all files in a directory or subdirectories.")
    category = "AAIT - TOOLBOX"
    icon = "icons/owfilesfromdir.svg"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/owfilesfromdir.svg"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owfindfilesfromdir.ui")
    want_control_area = False
    priority = 1060
    selected_column_name = Setting("input_dir")
    extension = Setting("")
    recursive = Setting("False")
    filter_per_date = Setting("False")
    extra_column = Setting("False")
    time_filter= Setting("0-0-0-1")

    class Inputs:
        data = Input("Data", Orange.data.Table)

    class Outputs:
        data = Output("Data", Orange.data.Table)


    @Inputs.data
    def set_path_table(self, in_data):
        self.data = in_data
        if in_data is None:
            self.Outputs.data.send(None)
            return
        if self.data:
            self.var_selector.add_variables(self.data.domain)
            self.var_selector.select_variable_by_name(self.selected_column_name)
        self.run()


    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(500)
        self.setFixedHeight(620)
        if self.is_dash_int_castable(self.time_filter)!=0:
            self.time_filter = "0-0-0-1"

        self.edit_extension = self.findChild(QLineEdit, 'lineEdit')
        self.edit_extension.setPlaceholderText("Extension (.docx, .pdf, .xslx, .csv, .json ...)")
        self.edit_extension.setText(self.extension)
        self.edit_extension.editingFinished.connect(self.update_parameters)
        self.comboBox = self.findChild(QCheckBox, 'checkBox')
        self.checkBox_2 = self.findChild(QCheckBox, 'checkBox_2')
        self.spinBox_day = self.findChild(QSpinBox, 'spinBox_day')
        self.spinBox_hour = self.findChild(QSpinBox, 'spinBox_hour')
        self.spinBox_minute = self.findChild(QSpinBox, 'spinBox_minute')
        self.spinBox_second= self.findChild(QSpinBox, 'spinBox_second')
        self.checkBox_extra_column = self.findChild(QCheckBox, 'checkBox_extra_column')

        DD,JJ,SS,MM=self._parse_delta_4_numbers(self.time_filter)
        self.spinBox_day.setValue(int(DD))
        self.spinBox_hour.setValue(int(JJ))
        self.spinBox_minute.setValue(int(SS))
        self.spinBox_second.setValue(int(MM))
        # Data Management
        self.folderpath = None
        self.data = None
        self.autorun = True
        self.post_initialized()

        if self.filter_per_date=="False":
            self.checkBox_2.setChecked(False)
            self.spinBox_day.setVisible(False)
            self.spinBox_hour.setVisible(False)
            self.spinBox_minute.setVisible(False)
            self.spinBox_second.setVisible(False)
        else:
            self.checkBox_2.setChecked(True)
            self.spinBox_day.setVisible(True)
            self.spinBox_hour.setVisible(True)
            self.spinBox_minute.setVisible(True)
            self.spinBox_second.setVisible(True)

        if self.recursive == "True":
            self.comboBox.setChecked(True)


        if self.extra_column=="True":
            self.checkBox_extra_column.setChecked(True)
        else:
            self.checkBox_extra_column.setChecked(False)

        self.comboBox.stateChanged.connect(self.on_checkbox_toggled)
        self.checkBox_2.stateChanged.connect(self.on_checkBox_2_toggled)
        self.checkBox_extra_column.stateChanged.connect(self.checkBox_extra_column_toggled)
        self.spinBox_day.valueChanged.connect(self.on_value_changed)
        self.spinBox_hour.valueChanged.connect(self.on_value_changed)
        self.spinBox_minute.valueChanged.connect(self.on_value_changed)
        self.spinBox_second.valueChanged.connect(self.on_value_changed)

        self.pushButton_run =self.findChild(QPushButton, 'pushButton_send')
        self.pushButton_run.clicked.connect(self.run)


    def on_value_changed(self):
        self.time_filter = str(self.spinBox_day.value())+"-"+str(self.spinBox_hour.value())+"-"+str(self.spinBox_minute.value())+"-"+str(self.spinBox_second.value())
    def update_parameters(self):
        self.extension = (self.edit_extension.text() or "").strip() #a jout de la gestion d'une zone vide
        if self.folderpath is not None:
            self.run()

    def on_checkbox_toggled(self,state):
        self.recursive = "True"
        if state==0:
            self.recursive = "False"
        if self.folderpath is not None:
            self.run()
    def on_checkBox_2_toggled(self,state):
        self.filter_per_date = "True"
        if state==0:
            self.filter_per_date = "False"

        if self.filter_per_date == "False":
            self.spinBox_day.setVisible(False)
            self.spinBox_hour.setVisible(False)
            self.spinBox_minute.setVisible(False)
            self.spinBox_second.setVisible(False)
        else:
            self.spinBox_day.setVisible(True)
            self.spinBox_hour.setVisible(True)
            self.spinBox_minute.setVisible(True)
            self.spinBox_second.setVisible(True)

        if self.folderpath is not None:
            self.run()

    def checkBox_extra_column_toggled(self,state):
        self.extra_column = "True"
        if state==0:
            self.extra_column = "False"



    def residual_path(self,parent_dir: str, child_path: str) -> str:
        """
        Renvoie la partie résiduelle de child_path par rapport à parent_dir.
        Ex: parent=C:/data, child=C:/data/a/b/file.txt => "a/b/file.txt"
        Indépendant des back slash et slash (Windows/Linux/mac).
        """
        parent = Path(parent_dir).expanduser().resolve()
        child = Path(child_path).expanduser().resolve()

        # Cas normal: child est dans parent
        try:
            rel = child.relative_to(parent)
            return rel.as_posix()  # format stable ("/")
        except ValueError:
            # Fallback: compare en "samefile" / normalisation, puis calcule un relpath
            # (utile si resolve() n’aligne pas tout comme prévu)
            import os
            rel2 = os.path.relpath(str(child), start=str(parent))
            return Path(rel2).as_posix()

    def find_files(self):
        files_data = []
        roots_data=[]
        files_data_rel=[]
        suffixes = self.parse_extensions()  # si pas de d'extensions, elles apparaitront toutes

        for i in range(len(self.folderpath)):
            base = self.folderpath[i]
            if self.recursive == "True":
                traversal = os.walk(base)
            else:
                try:
                    traversal = [(base, [], os.listdir(base))]
                except Exception:
                    continue

            for root, _, files in traversal:
                for file in files:
                    name = file.lower()
                    if suffixes is None or name.endswith(suffixes):
                        files_data.append([os.path.join(root, file).replace("\\", "/")])
                        roots_data.append(base)
                        files_data_rel.append(self.residual_path(base,os.path.join(root, file).replace("\\", "/")))

        return [files_data,roots_data,files_data_rel]

    def parse_extensions(self):
        """
        Convertit la saisie utilisateur en tuple de suffixes normalisés pour endswith.
        Exemples d'entrées valides :
          - ".pdf, .docx"
          - "pdf docx"
          - "csv"
          - "" (vide => aucune filtration, donc toutes extensions)
        """
        raw = (self.extension or "").strip().lower()
        if not raw:
            return None  # pas de filtre => tout passe

        # accepte virgules ou espaces multiples comme séparateurs
        parts = [p.strip() for chunk in raw.split(",") for p in chunk.split()]
        # normalise en ajoutant le point s'il manque, ignore les vides
        cleaned = []
        for p in parts:
            if not p:
                continue
            if not p.startswith("."):
                p = "." + p
            cleaned.append(p)

        if not cleaned:
            return None
        return tuple(set(cleaned))  # tuple unique pour endswith(...)

    def is_dash_int_castable(self,value: str) -> int:
        """
        Retourne 0 si tous les morceaux séparés par '-' sont castables en int, 1 sinon.
        Exemple :
          "1-0-0-1" -> 0
          "10-5-3-2" -> 0
          "1-a-0-1" -> 1
        """
        try:
            parts = value.split("-")
            if len(parts)!=4:
                return 1

            for p in parts:
                int(p)  # essaie de caster chaque partie
            return 0
        except (ValueError, TypeError):
            return 1

    def _parse_delta_4_numbers(self,delta_str):
        return(delta_str.split("-"))



    def _parse_delta(self,delta_str):
        """
        Accepte soit:
          - 'AA-MM-JJ-HH-mm-SS' (années, mois, jours, heures, minutes, secondes)
          - 'JJJJJ-HH-mm-SS'   (jours, heures, minutes, secondes)
        Les valeurs peuvent être non paddées (ex: '1-0-0-1').
        """
        parts = delta_str.split("-")
        try:
            nums = [int(p) for p in parts]
        except Exception as e:
            raise ValueError(
                "Delta invalide '{}'. Parties non entières.".format(delta_str)
            ) from e

        if len(nums) == 6:
            aa, mm, jj, hh, mi, ss = nums
            days = aa * 365 + mm * 30 + jj  # approximation mois=30j, année=365j
            return timedelta(days=days, hours=hh, minutes=mi, seconds=ss)
        elif len(nums) == 4:
            jj, hh, mi, ss = nums
            return timedelta(days=jj, hours=hh, minutes=mi, seconds=ss)
        else:
            raise ValueError(
                "Delta invalide '{}'. Formats acceptés: "
                "AA-MM-JJ-HH-mm-SS OU JJJJJ-HH-mm-SS.".format(delta_str)
            )

    def _created_or_modified_time(self,path):
        """
        Retourne l'instant (UTC) le plus récent entre création (si dispo) et modification.
        - Windows: st_ctime ~ création
        - Unix: st_ctime = change time; on utilise st_birthtime si disponible.
        """
        st = os.stat(path)
        mtime = st.st_mtime
        birth = getattr(st, "st_birthtime", None)
        best_epoch = max(mtime, birth) if birth is not None else mtime
        return datetime.utcfromtimestamp(best_epoch)

    def filter_files_newer_than_delta(self,
            files,
            delta_str,
            now=None,
            skip_missing=True,
    ):
        """
        Renvoie les fichiers créés/modifiés depuis moins que le delta donné.

        :param files: liste de chemins absolus OU chaîne avec chemins séparés.
        :param delta_str: 'AA-MM-JJ-HH-mm-SS' OU 'JJJJJ-HH-mm-SS' (sans padding requis).
        :param now: (optionnel) datetime de référence (UTC). Par défaut: datetime.utcnow().
        :param skip_missing: True -> ignore fichiers manquants, False -> lève FileNotFoundError.
        :return: liste de chemins satisfaisant (created/modified) >= now - delta
        """

        delta = self._parse_delta(delta_str)
        ref = now or datetime.utcnow()
        threshold = ref - delta
        results = []
        for pp in files:
            p=pp[0]

            if not os.path.isabs(p):
                # Chemins absolus attendus; on n'empêche pas pour autant.
                pass
            if not os.path.exists(p):
                if skip_missing:
                    continue
                raise FileNotFoundError(p)
            try:

                t = self._created_or_modified_time(p)
            except Exception:
                if not skip_missing:
                    raise
                continue
            if t >= threshold:
                results.append([p])
        return results

    def run(self):
        self.error("")
        self.warning("")

        if self.data is None:
            self.Outputs.data.send(None)
            return

        if not self.selected_column_name in self.data.domain:
            self.warning(f'Previously selected column "{self.selected_column_name}" does not exist in your data.')
            self.Outputs.data.send(None)
            return
        if self.extra_column=="True":
            if "root_dir" in  self.data.domain:
                self.error("The input column must not contain root_dir.")
                return
            if "relative_path" in  self.data.domain:
                self.error("The input column must not contain relative_path.")
                return

        self.folderpath = self.data.get_column(self.selected_column_name)

        try:
            files_data,roots_data,files_data_rel = self.find_files()

            if self.filter_per_date !="False":
                files_data=self.filter_files_newer_than_delta(files_data,self.time_filter)

            if len(files_data) == 0:
                self.Outputs.data.send(None)
                return

            if self.extra_column!="True":
                X = [[] for _ in files_data]
                domain = Domain([], metas=[StringVariable("path")])
                table = Table.from_numpy(domain, X, metas=files_data)
            else:
                def _scalar_str(v):
                    # Transforme v en string scalaire (si v est un tableau/list 1-élément, prend v[0])
                    if isinstance(v, (list, tuple, np.ndarray)):
                        if len(v) == 1:
                            v = v[0]
                    return "" if v is None else str(v)

                X = [[] for _ in range(len(files_data))]

                domain = Domain(
                    [],
                    metas=[
                        StringVariable("path"),
                        StringVariable("root_dir"),
                        StringVariable("relative_path"),
                    ],
                )

                metas = [
                    [_scalar_str(p), r, rel]
                    for p, r, rel in zip(files_data, roots_data, files_data_rel)
                ]

                table = Table.from_numpy(domain, X, metas=metas)
            self.Outputs.data.send(table)
        except Exception as e:
            self.error(f"An error occurred: the provided file path may not be supported ({e})")
            return

    def post_initialized(self):
        pass

if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWFileFromDir()
    my_widget.show()

    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()



import os
import sys
import Orange.data
from AnyQt.QtWidgets import QApplication
from Orange.widgets.utils.signals import Input, Output
from Orange.widgets.widget import OWWidget
import fitz  # PyMuPDF
from Orange.data import ContinuousVariable
import copy
if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
    from Orange.widgets.orangecontrib.AAIT.utils.import_uic import uic
else:
    from orangecontrib.AAIT.utils.import_uic import uic


class OWGetPages(OWWidget):
    name = "Get Pages"
    description = ("Extract the PDF page number corresponding to a text chunk contained in the document. The data table must contain two columns: the PDF path (path) and the text chunks (Chunks)")
    category = "AAIT - LLM INTEGRATION"
    icon = "icons/book.png"
    if "site-packages/Orange/widgets" in os.path.dirname(os.path.abspath(__file__)).replace("\\", "/"):
        icon = "icons_dev/book.png"
    gui = os.path.join(os.path.dirname(os.path.abspath(__file__)), "designer/owgetpages.ui")
    want_control_area = False
    priority = 1060

    class Inputs:
        data = Input("Data", Orange.data.Table)

    class Outputs:
        data = Output("Data", Orange.data.Table)


    @Inputs.data
    def set_path_table(self, in_data):
        self.data = in_data
        if in_data is None:
            self.Outputs.data.send(None)
            return
        self.run()

    def __init__(self):
        super().__init__()
        # Qt Management
        self.setFixedWidth(600)
        self.setFixedHeight(300)
        uic.loadUi(self.gui, self)
        self.data = None

    def load_pdf_with_sparse_mapping(self, pdf_path):
        """
        Load PDF thanks to fitz and create a mapping to identify pages limits.
        The pages containing the chunks will then be identified efficiently.

        :param pdf_path: The path to a pdf.
        :return: A dictionary containing the limit indexes for each page of the document.
        """
        # Load the pdf
        doc = fitz.open(pdf_path)
        full_text = ""
        page_mapping = {}  # Sparse mapping: {page_num: (start_index, end_index)}

        # Iterate over each page
        for page_num in range(len(doc)):
            # Get the text from the page
            page_text = doc[page_num].get_text()
            # Get the start index
            start_index = len(full_text)
            full_text += page_text
            # Get the end index
            end_index = len(full_text) - 1
            # Store the indexes for current page
            page_mapping[page_num + 1] = (start_index, end_index)

        doc.close()
        return full_text, page_mapping

    def find_pages_for_extract(self, full_text, page_mapping, extract):
        """
        Identify the pages that a given extract belongs to.

        :param full_text: The complete text of the PDF.
        :param page_mapping: A dictionary with page numbers as keys and (start_index, end_index) as values.
        :param extract: The text snippet to locate.
        :return: A list of page numbers the extract spans.
        """
        # Find the start index of the extract in the full text
        start_index = full_text.find(extract)
        if start_index == -1:
            return []  # Extract not found

        # Determine the end index of the extract
        end_index = start_index + len(extract) - 1

        # Find all pages the extract spans
        pages = []
        for page, (start, end) in page_mapping.items():
            if start <= end_index and end >= start_index:
                pages.append(page)

        return pages


    def run(self):
        self.error(None)
        if not "path" in self.data.domain:
            self.error('You don\'t have "path" column in your input data.')
            self.Outputs.data.send(None)
            return

        if not "Chunks" in self.data.domain:
            self.error('You don\'t have "Chunks" column in your input data.')
            self.Outputs.data.send(None)
            return
        pages = []
        for row in self.data:
            filepath = row["path"].value if os.path.isfile(row["path"].value) else os.path.join(row["path"].value,
                                                                                                row["name"].value)
            search_text = row["Chunks"].value
            try:
                full_text, page_mapping = self.load_pdf_with_sparse_mapping(filepath)
                page = self.find_pages_for_extract(full_text, page_mapping, search_text)[0]
            except:
                page = 1
            pages.append(page)
        if pages:
            print(f"The text was found on page(s): {pages}")
        else:
            print("The text was not found in the PDF.")

        data = copy.deepcopy(self.data)
        data = data.add_column(ContinuousVariable("page"), pages)
        self.Outputs.data.send(data)

    def post_initialized(self):
        pass

if __name__ == "__main__":
    app = QApplication(sys.argv)
    my_widget = OWGetPages()
    my_widget.show()

    if hasattr(app, "exec"):
        app.exec()
    else:
        app.exec_()
